"""A creation workflow for adding an existing NREN L3 Core Service to the service database."""

from uuid import uuid4

from orchestrator import workflow
from orchestrator.forms import FormPage
from orchestrator.targets import Target
from orchestrator.types import FormGenerator, SubscriptionLifecycle
from orchestrator.workflow import StepList, begin, done, step
from orchestrator.workflows.steps import resync, set_status, store_process_subscription
from pydantic import BaseModel
from pydantic_forms.types import UUIDstr

from gso.products import ProductName
from gso.products.product_blocks.bgp_session import BGPSession, IPFamily, IPTypes
from gso.products.product_blocks.nren_l3_core_service import NRENAccessPortInactive
from gso.products.product_blocks.service_binding_port import BFDSettings, ServiceBindingPortInactive
from gso.products.product_types.edge_port import EdgePort
from gso.products.product_types.nren_l3_core_service import ImportedNRENL3CoreServiceInactive, NRENL3CoreServiceType
from gso.services.partners import get_partner_by_name
from gso.services.subscriptions import get_product_id_by_name
from gso.utils.shared_enums import SBPType
from gso.utils.types.ip_address import IPAddress, IPv4AddressType, IPV4Netmask, IPv6AddressType, IPV6Netmask
from gso.utils.types.virtual_identifiers import VLAN_ID


def initial_input_form_generator() -> FormGenerator:
    """Take all information passed to this workflow by the :term:`API` endpoint that was called."""

    class BFDSettingsModel(BaseModel):
        bfd_enabled: bool = False
        bfd_interval_rx: int | None = None
        bfd_interval_tx: int | None = None
        bfd_multiplier: int | None = None

    class BaseBGPPeer(BaseModel):
        bfd_enabled: bool = False
        has_custom_policies: bool = False
        authentication_key: str | None
        multipath_enabled: bool = False
        send_default_route: bool = False
        is_passive: bool = False
        peer_address: IPAddress
        families: list[IPFamily]
        is_multi_hop: bool
        rtbh_enabled: bool

    class ServiceBindingPort(BaseModel):
        edge_port: UUIDstr
        ap_type: str
        geant_sid: str
        sbp_type: SBPType = SBPType.L3
        is_tagged: bool = False
        vlan_id: VLAN_ID
        custom_firewall_filters: bool = False
        ipv4_address: IPv4AddressType
        ipv4_mask: IPV4Netmask
        ipv6_address: IPv6AddressType
        ipv6_mask: IPV6Netmask
        rtbh_enabled: bool = True
        is_multi_hop: bool = True
        bgp_peers: list[BaseBGPPeer]
        v4_bfd_settings: BFDSettingsModel
        v6_bfd_settings: BFDSettingsModel

    class ImportNRENL3CoreServiceForm(FormPage):
        partner: str
        service_binding_ports: list[ServiceBindingPort]
        service_type: NRENL3CoreServiceType

    user_input = yield ImportNRENL3CoreServiceForm

    return user_input.model_dump()


@step("Create subscription")
def create_subscription(partner: str, service_type: NRENL3CoreServiceType) -> dict:
    """Create a new subscription object in the database."""
    partner_id = get_partner_by_name(partner)["partner_id"]
    if service_type == NRENL3CoreServiceType.GEANT_IP:
        product_id = get_product_id_by_name(ProductName.IMPORTED_GEANT_IP)
    elif service_type == NRENL3CoreServiceType.IAS:
        product_id = get_product_id_by_name(ProductName.IMPORTED_IAS)
    subscription = ImportedNRENL3CoreServiceInactive.from_product_id(product_id, partner_id)
    return {"subscription": subscription, "subscription_id": subscription.subscription_id}


@step("Initialize subscription")
def initialize_subscription(subscription: ImportedNRENL3CoreServiceInactive, service_binding_ports: list) -> dict:
    """Initialize the subscription with the user input."""
    for service_binding_port in service_binding_ports:
        edge_port_subscription = EdgePort.from_subscription(service_binding_port.pop("edge_port"))
        bgp_peers = service_binding_port.pop("bgp_peers")
        sbp_bgp_session_list = [
            BGPSession.new(
                subscription_id=uuid4(),
                ip_type=IPTypes.IPV4
                if session["families"] in {IPFamily.V4UNICAST, IPFamily.V4MULTICAST}
                else IPTypes.IPV6,
                **session,
            )
            for session in bgp_peers
        ]
        service_binding_port_subscription = ServiceBindingPortInactive.new(
            subscription_id=uuid4(),
            edge_port=edge_port_subscription.edge_port,
            sbp_bgp_session_list=sbp_bgp_session_list,
            v4_bfd_settings=BFDSettings.new(subscription_id=uuid4(), **(service_binding_port.pop("v4_bfd_settings"))),
            v6_bfd_settings=BFDSettings.new(subscription_id=uuid4(), **(service_binding_port.pop("v6_bfd_settings"))),
            **service_binding_port,
        )
        subscription.nren_l3_core_service.nren_ap_list.append(
            NRENAccessPortInactive.new(
                subscription_id=uuid4(),
                ap_type=service_binding_port["ap_type"],
                sbp=service_binding_port_subscription,
            )
        )

    subscription.description = f"{subscription.product} service"

    return {"subscription": subscription}


@workflow(
    "Create imported NREN L3 Core Service",
    initial_input_form=initial_input_form_generator,
    target=Target.CREATE,
)
def create_imported_nren_l3_core_service() -> StepList:
    """Import a GÉANT IP without provisioning it."""
    return (
        begin
        >> create_subscription
        >> store_process_subscription(Target.CREATE)
        >> initialize_subscription
        >> set_status(SubscriptionLifecycle.ACTIVE)
        >> resync
        >> done
    )
