from unittest.mock import patch

import pytest
from orchestrator.types import SubscriptionLifecycle

from gso.products import ProductName
from gso.products.product_types.nren_l3_core_service import NRENL3CoreService
from gso.services.subscriptions import get_product_id_by_name
from gso.utils.shared_enums import APType
from test.workflows import assert_complete, assert_lso_interaction_success, extract_state, run_workflow


@pytest.fixture()
def base_bgp_peer_input(faker):
    def _base_bgp_peer_input():
        bfd_enabled = faker.boolean()
        return {
            "bfd_enabled": bfd_enabled,
            "has_custom_policies": faker.boolean(),
            "authentication_key": faker.password(),
            "multipath_enabled": faker.boolean(),
            "send_default_route": faker.boolean(),
            "is_passive": faker.boolean(),
        }

    return _base_bgp_peer_input


@pytest.mark.parametrize("l3_core_type", [ProductName.GEANT_IP, ProductName.IAS])
@pytest.mark.workflow()
@patch("gso.services.lso_client._send_request")
def test_create_nren_l3_core_service_success(
    mock_lso_client,
    l3_core_type,
    faker,
    partner_factory,
    edge_port_subscription_factory,
    base_bgp_peer_input,
    data_config_filename,
):
    partner = partner_factory()
    product_id = get_product_id_by_name(l3_core_type)
    edge_port_a = edge_port_subscription_factory(partner=partner)

    form_input_data = [
        {"product": product_id},
        {"tt_number": faker.tt_number(), "partner": partner["partner_id"]},
        {"edge_port": {"edge_port": edge_port_a, "ap_type": APType.PRIMARY}},
        {
            "geant_sid": faker.geant_sid(),
            "is_tagged": faker.boolean(),
            "vlan_id": faker.vlan_id(),
            "ipv4_address": faker.ipv4(),
            "ipv4_mask": faker.ipv4_netmask(),
            "ipv6_address": faker.ipv6(),
            "ipv6_mask": faker.ipv6_netmask(),
            "custom_firewall_filters": faker.boolean(),
            "v4_bfd_settings": {
                "bfd_enabled": True,
                "bfd_interval_rx": faker.pyint(),
                "bfd_interval_tx": faker.pyint(),
                "bfd_multiplier": faker.pyint(),
            },
            "v6_bfd_settings": {
                "bfd_enabled": True,
                "bfd_interval_rx": faker.pyint(),
                "bfd_interval_tx": faker.pyint(),
                "bfd_multiplier": faker.pyint(),
            },
            "v4_bgp_peer": base_bgp_peer_input() | {"add_v4_multicast": faker.boolean(), "peer_address": faker.ipv4()},
            "v6_bgp_peer": base_bgp_peer_input() | {"add_v6_multicast": faker.boolean(), "peer_address": faker.ipv6()},
        },
    ]
    lso_interaction_count = 6

    result, process_stat, step_log = run_workflow("create_nren_l3_core_service", form_input_data)

    for _ in range(lso_interaction_count):
        result, step_log = assert_lso_interaction_success(result, process_stat, step_log)

    assert_complete(result)
    state = extract_state(result)
    subscription = NRENL3CoreService.from_subscription(state["subscription_id"])
    assert mock_lso_client.call_count == lso_interaction_count
    assert subscription.status == SubscriptionLifecycle.ACTIVE
    assert len(subscription.nren_l3_core_service.nren_ap_list) == 1
    assert (
        str(subscription.nren_l3_core_service.nren_ap_list[0].sbp.edge_port.owner_subscription_id)
        == form_input_data[2]["edge_port"]["edge_port"]
    )
